{This unit contains all the routines needed for handling files with fixed
length (up to 65528 bytes) records.  Records are identified by non negative
longints and all of the routines refer to these numbers. To use any of the
routines just drop a TFFile component on a form, set its FFileName
property and call AttachFile to use the routines. (See below for
details). Use the Create FFile Wizard to create a FFile data set.}

unit FFileU;

{$A-} {pack records here}

interface

uses StdCtrls, ExtCtrls, Graphics, Forms, Buttons, Classes,
   Controls, Windows, Dialogs, SysUtils;

type
  TFFile = class (TComponent)
  private
    ActiveRecs:      longint;  {Number of active records}
    RecordLen:       word;
    DataF:           file;
    FFileName: string;
    FAttached:   boolean;
    FRecordLength: word;
    FRecordCount: longint;
    procedure SetFFileName(Value: string);
    procedure SetRecordLength(Value: word);
  published

    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure   AttachFile;
    procedure   DetachFile;
    function    Exist(ri: longint): boolean;
    function    Insert(ri: longint; Buf: pointer): boolean;
    function    Delete(ri: longint): boolean;
    function    Fetch(ri: longint; Buf: pointer): boolean;
    function    Restore(ri: longint; Buf: pointer): boolean;
    function    FetchPartial(ri: longint; Buf: pointer;
                    Start,Trans: word): boolean;
    function    RestorePartial(ri: longint; Buf: pointer;
                    Start,Trans: word): boolean;
    property    FileName: string
                read  FFileName
                write SetFFileName;
    property    RecordLength: word
                read FRecordLength
                write SetRecordLength;
    property    RecordCount: longint
                read FRecordCount;
    end;


{FileName:  This is the name of a data set that contains the actual records.
The data set must exixt on disk when this property is set since
SetFFileName reads it to set some internal fields.

RecordLength:  Just the length of records in the file. Read only, but
visible in Object Inspector.

RecordCount:  Just the number of records in the file. Read only.

AttachFile: This method hooks up the internal TFFile with the disk data
set named in the FFileName property. It MUST be called before any other
methods or they wont work. A good place to put this is in the OnCreate
event for the form.

DetachFile. This method need not be called unless you want to use the
TFFile on another data set. This method breaks the connection between
the data set and the TFFile. Here is an example. We wish to disconnect
the TFFile from the data set 'DATA1' and connect instead to 'DATA2'.

   AFFile.DetachFFile;  -- disconnects DATA1 --
   AFFile.FFileName := 'DATA2';
   AFFile.AttachFile;

Exist: Returns TRUE if the record with record number rid exists.

Fetch: This method gets the record with id rid and places it in the
buffer pointed to by Buf.

Insert: This method is used to insert a new record with id rid. The
record is stored is the buffer pointer to by Buf. This method checks
to see that the id rid is not being used already, and if it is,
 does nothing.

Restore: This method is used to restore a fetched record. The id is rid.
This method checks to see that the id rid is being used already,
and if it is not does nothing.

Delete: This method is used to delete a record. The id is rid. This
method check to see that the id rid is being used already, and it not,
does nothing. When a record is deleted its place in
file has its first 7 bytes set to 'DELETED'.


FetchPartial:  Like fetch but only part of the record is retrieved.
Transmission starts at Start and Trans bytes are transmitted.


RestorePartial: Like Restore but only part of the record is retrieved.
Is intended to work with FetchPartial. }

procedure Register;


implementation


procedure Register;
begin
  RegisterComponents('Len''s',[TFFile]);
  end;

type
  Header = record
    ActiveRecs:      longint;
    RecordLen:       word;
    end;


{The file header is always the first 6 bytes in the file.
 The fields are:
    ActiveRecs:  The number of records in the file not counting
       deleted ones. This may be smaller than the filesize.
    RecordLen:  The length of the records. Set by the CreateFFile
        Wizard. }


{============= Set property routines ==================}

procedure TFFile.SetFFileName(Value: string);

var
  H: Header;

begin
  if FFileName <> Value
  then begin
    if Value = ''
    then if FAttached
         then DetachFile
         else
    else begin
      AssignFile(DataF,Value);
      {$I-} reset(DataF,1); {$I-}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+Value+' does not exist.');
        exit;
        end;
      Blockread(DataF,H,6);
      FRecordLength := H.RecordLen;
      FRecordCount := H.ActiveRecs;
      Close(DataF);
      end;
    FFileName := Value;
    end;
  end;


procedure  TFFile.SetRecordLength(Value: word);

begin
  end;


{FFile methods}

constructor TFFile.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FFileName := '';
  FAttached := FALSE;
  end;


destructor TFFile.Destroy;

begin
  DetachFile;
  inherited Destroy;
  end;



procedure TFFile.AttachFile;

begin
  if FAttached
  then begin
    ShowMessage('File already attached');
    exit;
    end;
  FAttached := FALSE; {Assume the worst}
  if FFileName = ''
  then begin
    ShowMessage('No file named yet.');
    exit;
    end;
  AssignFile(DataF,FFileName);
  reset(DataF,1);
  Blockread(DataF,ActiveRecs,6);
  FRecordLength := RecordLen;
  FRecordCount := ActiveRecs;
  FAttached := TRUE;
  end;


procedure TFFile.DetachFile;

begin
  if not FAttached
  then exit;
  if FFileName <> ''
  then begin
    seek(DataF,0);
    BlockWrite(DataF,FRecordCount,4);
    CloseFile(DataF);
    end;
  FFileName := '';
  FAttached := FALSE;
  end;


function TFFile.Exist(ri: longint): boolean;

var
  Deleted: byte;

begin
  Result := TRUE; {Assume OK}
  {First check that the record id is smaller than the file size.}
  if ri*FRecordLength+6 >= filesize(DataF)
  then Result := FALSE
  else begin {See if it's deleted.}
    seek(DataF,ri*FRecordLength+6);
    BlockRead(DataF,Deleted,1);
    if Deleted = 255
    then Result := FALSE;
    end;
  end;




function TFFile.Insert(ri: longint; Buf: pointer): boolean;

begin
  Result := not Exist(ri);
  if Result
  then begin
    Seek(DataF,ri*FRecordLength+6);
    BlockWrite(DataF,Buf^,FRecordLength);
    inc(FRecordCount);
    end;
  end;


function TFFile.Fetch(ri: longint; Buf: pointer): boolean;

begin
  Result := Exist(ri);
  if Result
  then begin
    seek(DataF,ri*FRecordLength+6);
    BlockRead(DataF,Buf^,FRecordLength);
    end;
  end;


function TFFile.Restore(ri: longint; Buf: pointer): boolean;

begin
  Result := Exist(ri);
  if Result
  then begin
    seek(DataF,ri*FRecordLength+6);
    BlockWrite(DataF,Buf^,FRecordLength);
    end;
  end;


function TFFile.Delete(ri: longint): boolean;

var
  Deleted: byte;

begin
  Result := Exist(ri);
  If Result
  then begin
    Deleted := 255;
    Seek(DataF,ri*FRecordLength+6);
    BlockWrite(DataF,Deleted,1);
    dec(FRecordCount);
    end;
  end;


function TFFile.FetchPartial(ri: longint; Buf: pointer;
              Start,Trans: word): boolean;

begin
  Result := Exist(ri);
  if Result
  then begin
    seek(DataF,ri*FRecordLength+Start+6);
    BlockRead(DataF,Buf^,Trans);
    end;
  end;


function TFFile.RestorePartial(ri: longint; Buf: pointer;
               Start,Trans: word): boolean;

begin
  Result := Exist(ri);
  if Result
  then begin
    seek(DataF,ri*FRecordLength+Start+6);
    BlockWrite(DataF,Buf^,Trans);
    end;
  end;

  end.

{====================END OF FFILE MODULE ====================}


